import assert from "assert";
import { Request, requestSettings, Response, responseSettings, } from "@foxify/http";
import inject from "@foxify/inject";
import Router from "@foxify/router";
import * as proxyAddr from "proxy-addr";
import * as qs from "qs";
import serveStatic from "serve-static";
import Server from "./Server";
import * as utils from "./utils";
import { Engine } from "./view";
const SETTINGS = [
    "env",
    "url",
    "port",
    "workers",
    "etag",
    "https",
    "https.cert",
    "https.key",
    "x-powered-by",
    "routing.case-sensitive",
    "routing.ignore-trailing-slash",
    "routing.allow-unsafe-regex",
    "routing.max-param-length",
    "json.escape",
    "json.replacer",
    "json.spaces",
    "jsonp.callback",
    "subdomain.offset",
    "trust.proxy",
    "query.parser",
];
class Foxify extends Router {
    static static = serveStatic;
    _settings = {
        env: process.env.NODE_ENV ?? "development",
        url: process.env.APP_URL ?? "localhost",
        // eslint-disable-next-line @typescript-eslint/no-magic-numbers
        port: process.env.APP_PORT ? +process.env.APP_PORT : 3000,
        workers: process.env.WORKERS ? +process.env.WORKERS : 1,
        https: false,
        "x-powered-by": true,
        "routing.allow-unsafe-regex": false,
        "routing.case-sensitive": true,
        "routing.ignore-trailing-slash": false,
        "routing.max-param-length": 100,
        "json.escape": false,
        "jsonp.callback": "callback",
        "subdomain.offset": 2,
        "query.parser": qs.parse,
        // eslint-disable-next-line no-undefined
        "trust.proxy": undefined,
        // eslint-disable-next-line no-undefined
        etag: undefined,
    };
    _view;
    constructor() {
        super();
        this.set("etag", "weak");
        this.disable("trust.proxy");
    }
    static dotenv(path) {
        utils.assertType("path", "string", path);
        // eslint-disable-next-line @typescript-eslint/no-var-requires,@typescript-eslint/no-require-imports
        require("dotenv").config({ path });
    }
    disable(setting) {
        return this.set(setting, false);
    }
    disabled(setting) {
        return !this.setting(setting);
    }
    enable(setting) {
        return this.set(setting, true);
    }
    enabled(setting) {
        return !this.disabled(setting);
    }
    /**
     * Handle view
     * @param extension view template file extension
     * @param path the directory containing view templates
     * @param handler
     */
    engine(extension, path, handler) {
        this._view = new Engine(path, extension, handler);
        return this;
    }
    async inject(options) {
        assert(this.setting("env") === "test", "Inject only works on the testing environment");
        if (typeof options === "string")
            options = { url: options };
        requestSettings(this._settings);
        responseSettings({
            ...this._settings,
            view: this._view,
        });
        // TODO: fix typescript issues
        return inject(this.lookup.bind(this), {
            ...options,
            // eslint-disable-next-line @typescript-eslint/ban-ts-comment
            // @ts-expect-error
            Response,
            Request,
        });
    }
    set(setting, value) {
        switch (setting) {
            case "env":
                utils.assertType(setting, "string", value);
                break;
            case "url":
                utils.assertType(setting, "string", value);
                break;
            case "port":
                utils.assertNonNegInt(setting, value);
                break;
            case "workers":
                utils.assertPosInt(setting, value);
                break;
            case "etag":
                if (value == null || typeof value === "function")
                    break;
                // eslint-disable-next-line no-undefined
                if (value === false)
                    value = undefined;
                else if (value === true || value === "weak")
                    value = utils.createETagGenerator(true);
                else if (value === "strong")
                    value = utils.createETagGenerator(false);
                else
                    throw new TypeError(`Unknown value for ${setting} setting: ${value}`);
                break;
            case "https":
                utils.assertType(setting, "boolean", value);
                break;
            case "https.cert":
                utils.assertType(setting, "string", value);
                break;
            case "https.key":
                utils.assertType(setting, "string", value);
                break;
            case "x-powered-by":
                utils.assertType(setting, "boolean", value);
                break;
            case "routing.case-sensitive":
                utils.assertType(setting, "boolean", value);
                break;
            case "routing.ignore-trailing-slash":
                utils.assertType(setting, "boolean", value);
                break;
            case "routing.allow-unsafe-regex":
                utils.assertType(setting, "boolean", value);
                break;
            case "routing.max-param-length":
                utils.assertPosInt(setting, value);
                break;
            case "json.escape":
                utils.assertType(setting, "boolean", value);
                break;
            case "json.replacer":
                utils.assertType(setting, "function", value);
                break;
            case "json.spaces":
                utils.assertNonNegInt(setting, value);
                break;
            case "jsonp.callback":
                utils.assertType(setting, "string", value);
                break;
            case "subdomain.offset":
                utils.assertNonNegInt(setting, value);
                break;
            case "trust.proxy":
                if (value === true) {
                    value = (() => true);
                }
                else if (typeof value === "number") {
                    const num = value;
                    value = ((ip, i) => i < num);
                }
                else if (typeof value !== "function") {
                    if (typeof value === "string")
                        value = value.split(/ *, */);
                    value = proxyAddr.compile(value || []);
                }
                break;
            case "query.parser":
                if (typeof value === "function")
                    break;
                if (value === true || value === "simple")
                    value = qs.parse;
                else if (value === false)
                    value = (() => ({}));
                else if (value === "extended")
                    value = ((str) => qs.parse(str, { allowPrototypes: true }));
                else
                    throw new TypeError(`Unknown value for ${setting} setting: ${value}`);
                break;
            default:
                throw new TypeError(`Expected setting to be one of [${SETTINGS}], got ${setting}`);
        }
        this._settings[setting] = value;
        return this;
    }
    setting(setting) {
        assert(SETTINGS.includes(setting), `Expected setting to be one of [${SETTINGS}], got ${setting}`);
        return this._settings[setting];
    }
    start(callback) {
        if (callback != null)
            utils.assertType("callback", "function", callback);
        /* Set node env */
        process.env.NODE_ENV = this.setting("env");
        const server = new Server({
            ...this._settings,
            view: this._view,
        }, this.lookup.bind(this));
        return server.start(callback);
    }
}
export default Foxify;
//# sourceMappingURL=Foxify.js.map